"""Curenttly, to work with flamewok, you may need to import some of
this objects:

Menu
TextBox
ActionBox
Form
Field
"""

from flamewok.menu import Menu, TextBox, ActionBox
from flamewok.form import Form, Field
from flamewok.validators import check_type
from flamewok.helpers import clear
import flamewok

infos = f"""
flamewok

version {flamewok.__version__}
home page: https://github.com/byoso/flamewok

"""

if __name__ == "__main__":
    print(infos)
    # help(Menu)
    # help(TextBox)
    # help(ActionBox)
    # help(Form)
    # help(Field)
    # help(check_type)



from uuid import uuid4
from abc import ABC

from flamewok import settings


class Box(ABC):
    """Box classes must inherit from this"""
    def __init__(self, label=""):
        self.label = label
        self.id = str(uuid4())

    def __str__(self):
        return f"<Box| {self.label}>"

    def __repr__(self):
        return str(self)


class ActionBox(Box):
    """An ActionBox is an activable element of the menu"""
    choices = []

    def __init__(self, choice="", label="", func=None):
        super().__init__(label)
        self.choice = str(choice)
        self.func = func

    def __str__(self):
        return f"<ActionBox| {self.choice} : {self.label} to {self.func}>"


class TextBox(Box):
    """A TextBox in a non-activable element of the menu, in fact, strings.
    """
    def __init__(self, label):
        super().__init__(label)

    def __str__(self):
        return f"<TextBox| {self.label} >"


class Menu:
    """Central class of the menu system"""

    def __init__(
        self,
        inside_sep=settings.DEFAULT_INSIDE_SEP,
        sep=settings.DEFAULT_SEP,
        box_size=settings.DEFAULT_BOX_SIZE,
        line_length=settings.DEFAULT_LINE_LENGTH,
        prompt=settings.DEFAULT_MENU_PROMPT,
        error_message=settings.DEFAULT_MENU_ERROR,
            ):
        """boxes is a list of tuples of strings built like this:
        - the tuples (will be the dialog boxes) have 3 or more elements:
        the user's choice to enter,
        a label describing the action,
        1 or more ids returned if choosen by the user
        - some str to insert in the menu (ex: "\n")
        """
        self.inside_sep = inside_sep
        self.sep = sep
        self.line_length = line_length
        self.box_size = box_size
        self.prompt = prompt
        self.response = None
        self.error_message = error_message
        self.boxes = []
        self.body = ""

    def __str__(self):
        return f"<Menu | {len(self._choices)} choices>"

    def __repr__(self):
        return str(self)

    def _build(self):
        """"Builds the aspect of the menu to display"""
        # manage the number of dialog boxes per line
        self.body = ""
        choices = []
        box_number = self.line_length // self.box_size
        for el in self.boxes:
            if isinstance(el, TextBox):
                self.body += el.label
            else:
                if el.choice in choices:
                    self.boxes.remove(el)
                else:
                    if box_number == 0:
                        self.body += "\n"
                        box_number = self.line_length // self.box_size
                    box_number -= 1
                    box = ""
                    box += f"{el.choice}{self.inside_sep}{el.label}"
                    self.body += (
                        f"{box[:self.box_size]:<{self.box_size}}{self.sep}")
                    choices.append(el.choice)

    def _input(self):
        """manage the input"""
        choice = input(self.prompt)
        if choice in self._choices:
            self.response = choice
            self._go_callback(choice)
        else:
            print(self.error_message)
            return self.ask()

    @property
    def _choices(self):
        """returns a list of all the choices used in the menu"""
        choices = []
        for box in self.boxes:
            if isinstance(box, ActionBox):
                choices.append(box.choice)
        return choices

    def _go_callback(self, choice):
        """Calls the box's associated function"""
        box = self.get_action_box(choice)
        return box.func()

    def ask(self):
        """Displays the menu and the prompt"""
        self._build()
        print(self.body)
        return self._input()

    def get_action_box(self, choice):
        """Returns an ActionBox from self.boxes"""
        box = list(filter(lambda b: isinstance(
            b, ActionBox) and b.choice == choice, self.boxes))[0]
        return box

    def get_box(self, index):
        """"Return any box from self.boxes"""
        box = self.boxes[index]
        return box

    def add_boxes(self, boxes):
        """boxes here are :
        - a list of tuple(s):
        [(choice, label, function), ...]
        - or a string"""
        for box in boxes:
            self.add_box(box)

    def add_box(self, box):
        """create and add a box from:
        - a tuple:
        (choice, label, function)
        - a string"""
        if isinstance(box, type(str())):
            new_box = TextBox(box)
            self.boxes.append(new_box)
        else:
            choice = box[0]
            if str(choice) in self._choices:
                pass
            else:
                label = box[1]
                func = box[2]
                new_box = ActionBox(choice, label, func)
                self.boxes.append(new_box)

    def insert(self, index, box):
        """Inserts a box in the menu at the index"""
        self.boxes.insert(index, box)

    def set_boxes(self, boxes):
        """Remove a former box list and replace with a new one"""
        self.boxes = []
        self.add_boxes(boxes)



#! /usr/bin/env python3


"""This module provides some validators to use in the forms or else.
validators are functions that always return a boolean
"""


def check_type(value, a_type):
    """Takes 2 parameters : a value and a type (bool, int or float)
    The fonction returns a boolean.
    check_type does NOT convert the value of the input,
    that remains a string.
    """
    if a_type == bool:
        if value.lower() in ["true", "false"]:
            return True
        else:
            return False
    else:
        try:
            a_type(value)
        except ValueError:
            return False
        return True





import flamewok
from flamewok.cli import cli

infos = f"""
=============================  flamewok  ===============================

Very tiny framework to quickly build python terminal applications.

version {flamewok.__version__}
home page: https://github.com/byoso/flamewok
pypi: https://pypi.org/project/flamewok/

"""


def cmd():
    print(infos)


if __name__ == "__main__":
    cli.route(
        ("", cmd, "show this help"),
        ("-h", cmd, "show this help"),
        ("--help", cmd, "show this help"),
    )


#! /usr/bin/env python3

import json
import os

"""
Simple json file loader / saver

Example:

json_data = JsonFile('data.json')

data = {
    'app1': {'folder': 'here', 'desktop': 'here'},
    'app2': {'folder': 'there', 'desktop': 'by there'}
}


json_data.save(data)
data = json_data.load()
data['app3'] = {'folder': 'new folder', 'desktop': 'somewhere'}
json_data.save(data)
results = json_data.load()
print(results)
"""


class JsonFile:
    """Interface with a json file"""

    def __init__(self, file_path):
        self.file = file_path

    def save(self, data):
        with open(self.file, 'w') as file:
            json.dump(data, file)

    def load(self):
        if os.path.exists(self.file):
            with open(self.file, 'r') as file:
                data = json.load(file)
            return data
        else:
            return []



"""Curenttly, to work with flamewok, you may need to import some of
this objects:

Menu
TextBox
ActionBox
Form
Field
"""

from flamewok.menu import Menu, TextBox, ActionBox
from flamewok.form import Form, Field
from flamewok.validators import check_type
from flamewok.helpers import clear

__version__ = "1.0.5"




from flamewok import settings


class NoNameError(Exception):
    def __init__(self):
        message = "Field with no name error"
        super().__init__(message)


class Field:
    """a Form needs some Field objects
    only 'name' is a necessary argument.
    """
    def __init__(
        self, name=None, label="", validator=None,
            error_message=None):
        if name:
            self.name = name
            self.label = label
            self.validator = validator
            self.message = error_message
        else:
            raise NoNameError

    def __str__(self):
        return f"<Field | {self.name} -> {self.label}>"

    def __repr__(self):
        return str(self)


class Response:
    """Single empty class used for the Form response"""
    # TODO: __str__ method
    pass


class Form:
    """Give a list of fields
    fields are tuples made of 2 elements:
    - str(the key to register the answer)
    - str(the text to display)
    Optional key words :
    - prompt: str
    - error_message: str
    """
    def __init__(self, fields: tuple, prompt=settings.DEFAULT_FORM_PROMPT,
                 error_message=settings.DEFAULT_FIELD_ERROR):
        self.message = error_message
        self.prompt = prompt
        self.fields = []
        for field in fields:
            new_field = Field(*field)
            self.fields.append(new_field)

    def ask(self):
        response = Response()
        for field in self.fields:
            if field.message:
                message = field.message
            else:
                message = self.message
            checked = False
            while not checked:
                print(f"{field.label}")
                value = input(self.prompt)
                if field.validator is not None:
                    checked = field.validator(value)
                    if not checked:
                        print(message)
                else:
                    checked = True
            setattr(response, field.name, value)
        return response



"""
This script allows to create veeeery quicly a CLI

Example:

if __name__ == "__main__":
    cli.route(
        "This is a test, to try my new CLI toy\n",
        ("", main, "Launches the main programm"),
        ("-h", cli.help, "displays this help"),
        ("--help", cli.help, 'Idem'),
        ("see <int:a> <bool:> <str:> <float:>", see, "Show the arguments given in the CLI"),
        ("test 2", test2, "calls the test2 function"),
        ("multi <float:a> <float:b>", multi, "calculate arg_a x arg_b"),
        ("add <numbers>", add, "print the addition of the given numbers")
    )

Each route takes a path with or without parameters '<>', a callback
function that revieve the parameters if there are some, and an optionnal
description.

that's all !
note that cli.help is an auto-created help.
"""


import sys

from flamewok import color as c


class CLI_Error(Exception):
    pass


class CLI:
    def __call__(self):
        print(sys.argv)

    def route(self, *args):
        """
        CLI.route takes the place of an __init__ method,
        but is explicitly called.
        args are tuples like this:
        ("path", function)
        """
        self.routes = []
        self.description = "| ==========    CLI options    ===========\n"

        for arg in args:
            if isinstance(arg, str):
                self.description += "* " + arg + "\n"

            else:
                if arg[0] == "":
                    self.description += '| [No option]'
                    if len(arg) >= 3:
                        self.description += f' : {arg[2]}'
                    self.description += '\n'

                else:
                    self.description += f"| {arg[0]}"
                    if len(arg) >= 3:
                        self.description += f' : {arg[2]}'
                    self.description += '\n'

                self.set_route(arg)

        command = sys.argv
        self.decode_command(command)

    def set_route(self, arg):
        if type(arg) == str:
            self.description += arg
        elif type(arg) == tuple:
            actions = []
            arguments = []
            path = arg[0].strip()
            function = arg[1]
            if len(arg) >= 3:
                help = arg[2]
            else:
                help = None
            elems = path.strip().split(" ")
            now_parameters = False
            for elem in elems:
                if elem.startswith("<"):
                    now_parameters = True
                    if elem.startswith("<int:"):
                        arguments.append(int)
                    elif elem.startswith("<float:"):
                        arguments.append(float)
                    elif elem.startswith("<bool:"):
                        arguments.append(bool)
                    else:
                        arguments.append(str)
                else:
                    if now_parameters:
                        raise CLI_Error("actions must be set BEFORE parameters")
                    actions.append(elem)

            self.routes.append({
                'path': path.strip(),
                'actions': actions,
                'function': function,
                'arguments': arguments,
                'help': help,
                })

    def find_route(self, commands, routes, index=0):
        if len(commands) < index:
            route = self.find_only_params_route(commands)
            return route
        if len(commands) == 0:
            routes = [route for route in routes if route['path'] == '']
            if len(routes) != 1:
                print("Command not found")
                exit()
        if len(routes) == 1:
            return routes[0]

        best = []
        arg_only_route = False
        for route in routes:
            # multiple routes with only arguments is forbiden
            if route['path'].startswith("<"):
                if arg_only_route:
                    print(f"{c.warning}error: Multiple routes have only <arguments> with no action before{c.end}")
                    exit()
                arg_only_route = True

            if len(route['actions']) > index and len(commands) > index:
                if route['actions'][index] == commands[index]:
                    best.append(route)
        index += 1

        return self.find_route(commands, best, index)

    def find_only_params_route(self, commands):
        """When no route 'with action' is found, try to find one with
        only parameters expected"""
        routes = [route for route in self.routes if route['actions'] == []]
        if len(routes) == 1:
            route = routes[0]
            return route
        print("Command not found")
        exit()

    def decode_command(self, command):
        commands = command[1:]
        routes = self.routes[:]

        route = self.find_route(commands, routes)

        index = len(route['actions'])
        args = commands[index:]
        if len(args) <= len(route['arguments']):
            min = len(args)
            max = None
        else:
            min = len(route['arguments'])
            max = len(args)

        params = []
        for i in range(min):
            if route['arguments'][i] == bool:
                try:
                    params.append(route['arguments'][i](int(args[i])))
                except ValueError:
                    print(
                        f"{c.warning}Aborted !\n"
                        "'boolean' type expected (0 or 1), "
                        f"recieved : {args[i]}{c.end}"
                        )
                    exit()
            else:
                try:
                    params.append(route['arguments'][i](args[i]))
                except ValueError:
                    print(
                        f"{c.warning}Wrong command !\n"
                        f"'{route['arguments'][i].__name__}' type expected, "
                        f"recieved : {args[i]}{c.end}"
                        )
                    exit()

        if max is not None:
            params = [*params, *args[min:max]]

        return route['function'](*params)

    def help(self):
        print(self.description)


cli = CLI()



#! /usr/bin/env python3
# coding: utf-8

"""
This file only belongs to pypi setup, not with silly_db in itself.
REMINDER:
1- build
./setup.py sdist bdist_wheel
2- basic verifications
twine check dist/*
2.5- Deploy on testpypi (optionnal, site here : https://test.pypi.org/):
twine upload --repository testpypi dist/*
3- upload to PyPi
twine upload dist/*
"""

import pathlib
from setuptools import setup
from flamewok import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="flamewok",
    version=f"{__version__}",
    description="Python micro framework for terminal UI applications",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/byoso/flamewok",
    author="Vincent Fabre",
    author_email="peigne.plume@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    packages=[
        "flamewok",
        "flamewok.tools"
        ],
    include_package_data=True,
    # install_requires=[],
    entry_points={
        "console_scripts": [
            "flamewok=flamewok.cmd:cmd",
        ]
    },
    setup_requires=['wheel'],
)




"""Here are defined the default values used in flamewok"""


# Menu default settings

DEFAULT_INSIDE_SEP = ": "
DEFAULT_SEP = " | "
DEFAULT_BOX_SIZE = 20
DEFAULT_LINE_LENGTH = 80
DEFAULT_MENU_PROMPT = "?> "
DEFAULT_MENU_ERROR = "\n!! Wrong entry, you must enter a valid choice !!"
DEFAULT_FIELD_ERROR = "Unexpected value, please try again"


# Form default settings

DEFAULT_FORM_PROMPT = "> "



#! /usr/bin/env python3
# coding: utf-8


"""This module provides some helpers to handle color in terminal

WARNING: the display of the colors is quite hasardous depending on the
terminal you use, and the os you are running. It seems ok in simple line
displayed string, but becomes very messy if used in a menu.

To change a color of a text, use a color mark in the str itself, and the end
mark to get back to normal.

To see a palette, import the function palette from the module,
or execute the module in a terminal:
    python3 -m flamewok.color

ex: import color as c

    title = c.Color(1, 45 ,37)
    print(f"{title.mark}A title{title.end}")

a few marks are ready to use in this module:

end = '\x1b[0m'

info = Color(0, 46, 30).mark
success = Color(0, 42, 30).mark
warning = Color(1, 43, 30).mark
danger = Color(1, 41, 33).mark

"""


class Color:
    """Color object, it requires 3 parameters (int):
    - style: from 0 to 8
    - background: from 0 to 50
    - foreground: from 0 to 50
    """
    def __init__(self, style, background, foreground,):
        self.style = style
        self.foreground = foreground
        self.background = background
        self.ref = (
            f"{str(self.style)};"
            f"{str(self.background)};"
            f"{str(self.foreground)}")
        self.mark = f"\x1b[{self.ref}m"
        self.end = "\x1b[0m"

    def __str__(self):
        return f"{self.mark} {self.ref} {end}"

    def __repr__(self):
        return f"<Color | {str(self)}>"


def palette(style_range=(0, 8), bg_range=(30, 38), fg_range=(40, 48)):
    """
    Showes some color references
    """
    for style in range(style_range[0], style_range[1]):
        for background in range(bg_range[0], bg_range[1]):
            show = ''
            for foreground in range(fg_range[0], fg_range[1]):
                color = Color(style, background, foreground)
                show += str(color)
            print(show)
        print('\n')


# Some prebuilt marks
end = "\x1b[0m"

info_hl = Color(0, 46, 30).mark
success_hl = Color(0, 42, 30).mark
warning_hl = Color(1, 43, 30).mark
danger_hl = Color(1, 41, 33).mark

info = Color(0, 30, 36).mark
success = Color(0, 30, 32).mark
warning = Color(0, 30, 33).mark
danger = Color(0, 30, 31).mark


if __name__ == "__main__":

    palette()

    print(f"{info} info {end}")
    print(f"{success} success {end}")
    print(f"{warning} warning {end}")
    print(f"{danger} danger {end}")
